# Funkce pro počítání s časem a daty.
# (c) 2006-2008 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

package cas;
require 5.000;
require Exporter;
use utf8;

@ISA = qw(Exporter);
@EXPORT = qw(esek2hash datumcas2esek esek2datumcas datum2eden eden2datum eden2esek esek2mcas sek2cas den_v_tydnu ted);



# Hlavní datové typy, které tento modul používá:
# datum: řetězec ve tvaru "8.5.2006"
# cas:   řetězec ve tvaru "15:37:56"
# datumcas: řetězec ve tvaru "8.5.2006 15:37:56"
# rmdhms: řetězec ve tvaru "20060508153756"
# mcas:  místní čas - stejný formát jako datumcas, ale hodnota je upravená podle časového pásma a letního času
# eden:  počet dní od začátku epochy (1. ledna v roce epochy je den č. 0), např. pro datum "8.5.2006" a epochu 1970 je to 13277
# esek:  počet nepřestupných sekund od začátku epochy, např. pro 8.5.2006 0:00:00 UTC a epochu 1970 je to 1147132800
# time:  počet nepřestupných sekund od začátku systémové epochy (pro MacOS 1904, pro ostatní 1970)
# hash:  kompletní soubor údajů odvoditelných pro jeden bod v čase (např. samostatně den, měsíc, rok atd.)



#------------------------------------------------------------------------------
# Nastaví epochu pro přepočítávání data na celé číslo a zpět. Epocha je číslo
# roku, tj. epocha začíná 1. ledna v tomto roce. (1. leden roku epochy je
# epochální den č. 0. Den č. 1 je 2. leden.)
#------------------------------------------------------------------------------
$epocha = 1970; # default
sub nastavit_epochu
{
    my $dosavadni_epocha = $epocha;
    $epocha = shift;
    return $dosavadni_epocha;
}



#------------------------------------------------------------------------------
# Zjistí epochu pro přepočítávání data na celé číslo a zpět. Epocha je číslo
# roku, tj. epocha začíná 1. ledna v tomto roce. (1. leden roku epochy je
# epochální den č. 0. Den č. 1 je 2. leden.)
#------------------------------------------------------------------------------
sub epocha
{
    return $epocha;
}



#------------------------------------------------------------------------------
# Zjistí pro libovolné datum, kolikátý je to den od začátku epochy (nastavena
# na začátku funkce, např. na 1.1.1970). Umožňuje provádět aritmetické operace
# s daty. 1. leden roku epochy je den č. 0.
#------------------------------------------------------------------------------
sub datum2eden
{
    my $epocha = epocha();
    my $datum = shift; # dd.mm.rrrr
    my $vysledek = 0;
    $datum =~ m/(\d+)\.(\d+)\.(\d+)/;
    my $den = $1;
    my $mesic = $2;
    my $rok = $3;
    my($irok, $imesic);
    my @delka_mesice = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
    # Postup se liší podle toho, zda se zadané datum nachází v epoše, nebo před ní.
    if($rok>=$epocha)
    {
        for($irok = $epocha; $irok<$rok; $irok++)
        {
            $vysledek += pocet_dni_v_roce($irok);
        }
        for($imesic = 1; $imesic<$mesic; $imesic++)
        {
            $vysledek += $delka_mesice[$imesic-1];
        }
        if(($mesic>2) && je_prestupny_rok($rok))
        {
            $vysledek++;
        }
        $vysledek += $den;
        # První den epochy má mít číslo 0, ale zatím jsme se chovali, jako by to měla být jednička. Odečíst ji.
        $vysledek--;
    }
    # Pro roky před epochou musíme postupovat v opačném směru.
    else # $rok<$epocha
    {
        for($irok = $epocha-1; $irok>$rok; $irok--)
        {
            $vysledek -= pocet_dni_v_roce($irok);
        }
        for($imesic = 12; $imesic>$mesic; $imesic--)
        {
            $vysledek -= $delka_mesice[$imesic-1];
        }
        if(($mesic<2) && je_prestupny_rok($rok))
        {
            $vysledek--;
        }
        $vysledek -= $delka_mesice[$mesic-1]-$den+1;
    }
    return $vysledek;
}



#------------------------------------------------------------------------------
# Přepočítá počet dnů od začátku epochy na datum ve formátu d.m.rrrr.
#------------------------------------------------------------------------------
sub eden2datum
{
    my $epocha = epocha();
    my $den = shift;
    my $rok;
    my $dni_v_roce;
    my $mesic;
    # Záporné dny (před epochou) se budou počítat jinak než kladné (v epoše).
    if($den>=0)
    {
        # První den epochy má číslo 0. Aby se nám lépe počítalo, posunout nejdříve o 1 nahoru.
        $den++;
        # Zjistit rok.
        $rok = $epocha;
        for(;;)
        {
            # Je tento rok přestupný?
            $dni_v_roce = pocet_dni_v_roce($rok);
            if($den>$dni_v_roce)
            {
                $den -= $dni_v_roce;
                $rok++;
            }
            else
            {
                last;
            }
        }
        # Zjistit měsíc.
        my $unor = $dni_v_roce==365 ? 28 : 29;
        my @delka_mesice = (31, $unor, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
        for($mesic = 1; $mesic<=12; $mesic++)
        {
            if($den>$delka_mesice[$mesic-1])
            {
                $den -= $delka_mesice[$mesic-1];
            }
            else
            {
                last;
            }
        }
    }
    # Záporné dny (před epochou) se budou počítat jinak než kladné (v epoše).
    else # $den<0
    {
        # Zjistit rok.
        $rok = $epocha-1;
        for(;;)
        {
            # Je tento rok přestupný?
            $dni_v_roce = pocet_dni_v_roce($rok);
            if($den<-$dni_v_roce)
            {
                $den += $dni_v_roce;
                $rok--;
            }
            else
            {
                last;
            }
        }
        # Zjistit měsíc.
        my $unor = $dni_v_roce==365 ? 28 : 29;
        my @delka_mesice = (31, $unor, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
        for($mesic = 12; $mesic>=1; $mesic--)
        {
            if($den<-$delka_mesice[$mesic-1])
            {
                $den += $delka_mesice[$mesic-1];
            }
            else
            {
                last;
            }
        }
        # Zjistit den.
        $den = $delka_mesice[$mesic-1]+$den+1;
    }
    my $datum = "$den.$mesic.$rok";
    return $datum;
}



#------------------------------------------------------------------------------
# Zjistí počet dní v roce v závislosti na tom, zda je to přestupný rok.
#------------------------------------------------------------------------------
sub pocet_dni_v_roce
{
    my $rok = shift;
    return je_prestupny_rok($rok) ? 366 : 365;
}



#------------------------------------------------------------------------------
# Zjistí, zda jde o přestupný, nebo normální rok. Nebere v úvahu, že přestupné
# roky byly zavedeny až s gregoriánským kalendářem v 16. století. Klidně ohlásí
# přestupný rok před Kristem.
#------------------------------------------------------------------------------
sub je_prestupny_rok
{
    my $rok = shift;
    if(($rok % 400 == 0) || ($rok % 4 == 0) && ($rok % 100 != 0))
    {
        return 1;
    }
    else
    {
        return 0;
    }
}



#------------------------------------------------------------------------------
# Pro zadané číslo dne od začátku epochy zjistí, kolikátý je to den v týdnu.
# (0 = neděle, 6 = sobota)
# Týden nám začíná po anglicku nedělí, abychom se zbytečně neodchylovali od
# toho, jak má dny číslované operační systém. Na druhou stranu díky tomu
# ostatní dny mají intuitivnější čísla (pondělí 1, pátek 5 atd.)
#------------------------------------------------------------------------------
sub den_v_tydnu
{
    my $den = shift;
    # Počítání dne v týdnu by mělo být nezávislé na epoše.
    # Co když někoho napadne použít jinou než defaultní epochu?
    # 1.1.2007 bylo pondělí, tj. 31.12.2006 byla neděle.
    my $den0 = datum2eden('31.12.2006');
    # Modulo bude správně fungovat i pro dny před "dnem 0":
    # -1 % 7 == 6
    # -6 % 7 == 1
    # -7 % 7 == 0
    # -8 % 7 == 6
    return ($den-$den0)%7;
}



#------------------------------------------------------------------------------
# Přepočítá počet sekund od začátku systémové epochy na počet sekund od začátku
# naší epochy.
#------------------------------------------------------------------------------
sub time2esek
{
    my $time = shift;
    # Pokud jsme tuto funkci již někdy volali, pak známe posun naší epochy vůči systémové.
    # V opačném případě ho musíme vypočítat.
    if($posun_epochy eq "")
    {
        # Nechat systém převést $time podle jeho epochy na univerzálně platné časové údaje.
        my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday) = gmtime($time);
        # Z těchto údajů vypočítat počet sekund od začátku naší epochy.
        my $esek = datum2eden("$mday.".($mon+1).".".($year+1900))*86400+$hour*3600+$min*60+$sec;
        # Abychom nemuseli tento výpočet opakovat, zapamatovat si v globální proměnné rozdíl mezi naší a systémovou epochou.
        $posun_epochy = $esek-$time;
    }
    return $time+$posun_epochy;
}



#------------------------------------------------------------------------------
# Přepočítá počet sekund od začátku naší epochy na počet sekund od začátku
# systémové epochy.
#------------------------------------------------------------------------------
sub esek2time
{
    my $esek = shift;
    # Pokud jsme tuto funkci již někdy volali, pak známe posun naší epochy vůči systémové.
    # V opačném případě ho musíme vypočítat.
    if($posun_epochy eq "")
    {
        # Perl umožňuje získat datum a čas na základě počtu sekund od začátku systémové epochy,
        # ale neumožňuje získat počet sekund od začátku systémové epochy na základě data a času.
        # Vzhledem k tomu, že neznáme systémovou epochu (nevíme, zda neběžíme pod MacOS nebo
        # něčím ještě kurióznějším), musíme si nechat systémem přepočítat nějaký čas (třeba 0)
        # na jednotky, ze kterých už posun dokážeme zjistit.
        time2esek(0);
    }
    return $esek-$posun_epochy;
}



#------------------------------------------------------------------------------
# Převezme počet nepřestupných sekund od začátku naší epochy a vrátí hash
# s různými údaji odvoditelnými pro daný časový bod.
#   sek ...... sekunda (0..59)
#   min ...... minuta (0..59)
#   hod ...... hodina (0..23)
#   den ...... den v měsíci (1..31)
#   mes ...... číslo měsíce (1..12)
#   rok ...... rok gregoriánského kalendáře (nenulové celé číslo)
#   imes ..... index měsíce pro vyhledávání v poli názvů měsíců (0..11)
#   idvt ..... index dne v týdnu pro vyhledávání v poli názvů dnů (0..6; 0 je neděle, 1 je pondělí!)
#   letni .... true, jestliže je právě letní čas (0..1)
#   eden ..... počet dní od začátku naší epochy (celé číslo)
#   esek ..... počet sekund od začátku naší epochy (celé číslo)
#   system ... počet sekund od začátku systémové epochy (celé číslo)
#   mesic .... název měsíce (leden..prosinec)
#   mesice ... název měsíce v genitivu (ledna..prosince)
#   dvt ...... název dne v týdnu (pondělí..neděle)
#   datum .... celé datum, např. "21.12.1971"
#   cas ...... celý čas, např. "15:37:05"
#   rmdhms ... datum a čas, který je možné přečíst (na rozdíl od esek), ale lze podle něj třídit, např. "19711221153705"
#------------------------------------------------------------------------------
sub esek2hash
{
    my $esek = shift;
    my $time = esek2time($esek);
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime($time);
    my %hash =
    (
        "sek" => $sec,   # sekunda
        "min" => $min,   # minuta
        "hod" => $hour,  # hodina
        "den" => $mday,  # den (číslo dne v měsíci)
        "mes" => $mon+1, # měsíc (číslo měsíce v roce)
        "rok" => 1900+$year, # rok
        "imes" => $mon,  # index měsíce pro vyhledávání v poli názvů měsíců
        "idvt" => $wday, # index dne v týdnu pro vyhledávání v poli názvů dnů (0 je neděle, 1 je pondělí)
        "letni" => $isdst, # true, jestliže je právě letní čas
        "esek" => $esek,   # počet sekund od začátku naší epochy
        "system" => $time, # počet sekund od začátku systémové epochy
    );
    $hash{mesic} = ("leden", "únor", "březen", "duben", "květen", "červen", "červenec", "srpen", "září", "říjen", "listopad", "prosinec")[$hash{imes}];
    $hash{mesice} = ("ledna", "února", "března", "dubna", "května", "června", "července", "srpna", "září", "října", "listopadu", "prosince")[$hash{imes}];
    $hash{dvt} = ("neděle", "pondělí", "úterý", "středa", "čtvrtek", "pátek", "sobota")[$hash{idvt}];
    $hash{datum} = sprintf("%d.%d.%d", $hash{den}, $hash{mes}, $hash{rok});
    $hash{cas} = sprintf("%d:%02d:%02d", $hash{hod}, $hash{min}, $hash{sek});
    $hash{rmdhms} = sprintf("%04d%02d%02d%02d%02d%02d", $hash{rok}, $hash{mes}, $hash{den}, $hash{hod}, $hash{min}, $hash{sek});
    # Předpočítat počet dnů od začátku mé epochy.
    $hash{eden} = datum2eden($hash{datum});
    return \%hash;
}



#------------------------------------------------------------------------------
# Zjistí pro libovolné datum a čas, kolik sekund uplynulo od začátku epochy.
#------------------------------------------------------------------------------
sub datumcas2esek
{
    my $datumcas = shift; # dd.mm.rrrr hh:mm:ss
    $datumcas =~ m/(\d+\.\d+\.\d+)\s+(\d+):(\d+):(\d+)/;
    my $datum = $1;
    my $hod = $2;
    my $min = $3;
    my $sek = $4;
    return datum2eden($datum)*86400+$hod*3600+$min*60+$sek;
}



#------------------------------------------------------------------------------
# Převede počet sekund od začátku epochy na datum a čas (UTC).
#------------------------------------------------------------------------------
sub esek2datumcas
{
    my $esek = shift;
    my $zarovnat = shift;
    my $eden = int($esek/86400);
    my $hod = int(($esek%86400)/3600);
    my $min = int(($esek%3600)/60);
    my $sek = $esek%60;
    my $datum = eden2datum($eden);
    my $nmd = $zarovnat ? "10" : "";
    my $nmh = $zarovnat ? "2" : "";
    return sprintf("%${nmd}s %${nmh}d:%02d:%02d", $datum, $hod, $min, $sek);
}



#------------------------------------------------------------------------------
# Převede datum a čas ve formátu rmdhms na počet sekund od začátku epochy.
#------------------------------------------------------------------------------
sub rmdhms2esek
{
    my $rmdhms = shift; # rrrrmmddhhmmss
    $rmdhms =~ m/(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)/;
    my $datum = sprintf("%d.%d.%d", $3, $2, $1);
    my $hod = $4;
    my $min = $5;
    my $sek = $6;
    return datum2eden($datum)*86400+$hod*3600+$min*60+$sek;
}



#------------------------------------------------------------------------------
# Převede počet sekund od začátku epochy na datum a čas (UTC) ve formátu rmdhms.
#------------------------------------------------------------------------------
sub esek2rmdhms
{
    my $esek = shift;
    my $eden = int($esek/86400);
    my $hod = int(($esek%86400)/3600);
    my $min = int(($esek%3600)/60);
    my $sek = $esek%60;
    my $datum = eden2datum($eden);
    $datum =~ m/(\d+)\.(\d+)\.(\d+)/;
    return sprintf("%d%02d%02d%02d%02d%02d", $3, $2, $1, $hod, $min, $sek);
}



#------------------------------------------------------------------------------
# Převede dny na sekundy.
#------------------------------------------------------------------------------
sub eden2esek
{
    my $eden = shift;
    return $eden*86400;
}



#------------------------------------------------------------------------------
# Převede počet sekund od začátku epochy na datum a čas v určitém časovém
# pásmu.
#------------------------------------------------------------------------------
sub esek2mcas
{
    my $esek = shift;
    my $pasmo = shift; # číslo od -12 do +12, udávající časový posun pásma oproti UTC
    my $uvest_pasmo = shift;
    my $zarovnat = shift;
    my $datumcas = esek2datumcas($esek+$pasmo*3600, $zarovnat);
    if($uvest_pasmo)
    {
        # Některé země nejsou od Greenwiche posunuté o celé hodiny.
        my $uprpasmo;
        my $desetiny = abs($pasmo)-int(abs($pasmo));
        if($desetiny)
        {
            $uprpasmo = sprintf("%+d:%d", $pasmo, 60*$desetiny);
        }
        else
        {
            $uprpasmo = sprintf("%+d", $pasmo);
        }
        $datumcas .= " GMT$uprpasmo";
    }
    return $datumcas;
}



#------------------------------------------------------------------------------
# Převede počet vteřin na větší časové jednotky. Sám si vybere největší
# takovou jednotku, jejíž počet ještě vyjde nenulový.
#------------------------------------------------------------------------------
sub sek2cas
{
    my $cas = shift;
    my $format = shift; # 0: bez jednotek, vždy v hodinách; 1: "7:50 m"; 2: "3 vteřiny"
    my $oddelovac = shift; # co vložit mezi číslo a jednotky; default je mezera; prázdný řetězec nejde zadat
    $oddelovac = " " if($oddelovac eq "");
    my $hod = int($cas/3600);
    my $min = int(($cas%3600)/60);
    my $sek = $cas%60;
    my $hlaseni;
    if($hod==0)
    {
        if($min==0)
        {
            if($format==2)
            {
                $hlaseni = sprintf("$sek${oddelovac}vteřin%s", $sek==1 ? "u" : $sek>=2 && $sek<=4 ? "y" : "");
            }
            else
            {
                $hlaseni = "$sek s";
            }
        }
        else
        {
            if($format==2)
            {
                $hlaseni = sprintf("%d:%02d${oddelovac}minut", $min, $sek);
            }
            else
            {
                $hlaseni = sprintf("%d:%02d${oddelovac}m", $min, $sek);
            }
        }
    }
    else
    {
        if($format==2)
        {
            $hlaseni = sprintf("%d:%02d:%02d${oddelovac}hodin", $hod, $min, $sek);
        }
        else
        {
            $hlaseni = sprintf("%d:%02d:%02d${oddelovac}h", $hod, $min, $sek);
        }
    }
    return $hlaseni;
}



#------------------------------------------------------------------------------
# Zjistí aktuální čas a vrátí odkaz na hash, ve kterém je aktuální čas různě
# formátovaný.
#------------------------------------------------------------------------------
sub ted
{
    # Zjistit počet nepřestupných sekund od začátku epochy (pozor, v tomto
    # případě nejde o moji epochu nastavenou v tomto modulu, nýbrž o systémovou
    # epochu - pro většinu systémů je to 1.1.1970 0:00:00 UTC, pro MacOS je to
    # 1.1.1904 0:00:00 UTC).
    my $time = time();
    return esek2hash(time2esek($time));
}



#------------------------------------------------------------------------------
# Vypíše dobu, po kterou program běžel. K tomu potřebuje dostat časové otisky
# začátku a konce.
#------------------------------------------------------------------------------
sub sestavit_hlaseni_o_trvani_programu
{
    my $starttime = shift;
    my $stoptime = time();
    my $cas = $stoptime-$starttime;
    my $hod = int($cas/3600);
    my $min = int(($cas%3600)/60);
    my $sek = $cas%60;
    my $hlaseni;
    if($hod==0)
    {
        if($min==0)
        {
            $hlaseni = sprintf("Program běžel $sek vteřin%s.\n", $sek==1 ? "u" : $sek>=2 && $sek<=4 ? "y" : "");
        }
        else
        {
            $hlaseni = sprintf("Program běžel %d:%02d minut.\n", $min, $sek);
        }
    }
    else
    {
        $hlaseni = sprintf("Program běžel %2d:%02d:%02d hodin.\n", $hod, $min, $sek);
    }
    return $hlaseni;
}
#------------------------------------------------------------------------------
# Kvůli zpětné kompatibilitě ještě tatáž funkce pod jiným jménem. Navzdory
# jménu nic nikam nevypisuje, pouze sestaví hlášení, které si volající vypíše
# sám, kam uzná.
#------------------------------------------------------------------------------
sub vypsat_delku_trvani_programu
{
    return sestavit_hlaseni_o_trvani_programu(@_);
}



#------------------------------------------------------------------------------
# Zjistí, kdy byl naposledy měněn soubor, ve vteřinách od začátku epochy (tady
# ovšem nejde o mou epochu z funkce zert::epocha(), ale o systémovou epochu).
#------------------------------------------------------------------------------
sub cassoubor
{
    my $soubor = $_[0];
    my ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, $ctime, $blksize, $blocks) = stat($soubor);
    # Není mi jasné, zda mám použít ctime (inode change time), nebo mtime (last modify time).
    return $mtime;
}



#------------------------------------------------------------------------------
# Gaussův algoritmus pro výpočet data velikonoční neděle v daném roce. Údajně
# platí pro všechny roky od roku 1900 do roku 2099 s výjimkou let 1954 a 1981,
# kdy byly velikonoce o týden dřív.
#------------------------------------------------------------------------------
sub velikonocni_nedele
{
    my $rok = shift;
    my $zbytek1 = $rok % 19;
    my $zbytek2 = $rok % 4;
    my $zbytek3 = $rok % 7;
    my $zbytek4 = ($zbytek1*19+24) % 30;
    my $zbytek5 = (5+2*$zbytek2+4*$zbytek3+6*$zbytek4) % 7;
    my $vysledek = datum2eden("22.3.$rok")+$zbytek4+$zbytek5;
    # Nevím, jestli nemám nějakou chybu v přepočítávání dat na epochy, protože eden2datum($vysledek) mi vrací vždy bílou sobotu.
    # Proto provizorní úprava.
    $vysledek++;
    return $vysledek;
}



1;
