#!/usr/bin/perl
# Knihovna funkcí pro práci se systémem souborů a pro bezpečné volání externích procesů v Linuxu.
# © 2010 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

package dzsys;
use utf8;
use open ":utf8";



#------------------------------------------------------------------------------
# Zařídí pro otevřený výstupní soubor autoflush mezi každými dvěma perlovými
# příkazy. Vhodné např. k tomu, aby se v terminálovém okně míchaly STDOUT a
# STDERR chronologicky.
#------------------------------------------------------------------------------
sub autoflush
{
    my $soubor = shift; # např. *STDOUT
    my $old_fh = select($soubor);
    $| = 1;
    select($old_fh);
}



#------------------------------------------------------------------------------
# Otevře soubor na disku ke čtení. Jestliže je soubor zagzipovaný, funkce ho
# při čtení rozgzipuje. Vrátí handle souboru. Hodí výjimku, jestliže soubor
# nejde otevřít (chybná cesta, nedostatečná práva atd.)
#------------------------------------------------------------------------------
sub gopen
{
    my $cesta = shift;
    if($cesta =~ m/\.gz$/)
    {
        $cesta = "gunzip -c $cesta |";
    }
    # Jestliže open() dostane místo handlu neinicializovanou lokální skalární
    # proměnnou, vyrobí a vloží do ní anonymní lokální handle, který sám zanikne,
    # až skončí aktuální blok a zaniknou všechny odkazy na něj. (Pokud tedy
    # volající nechce, aby mu handle zanikl, musí ho vložit do globální proměnné.)
    my $handle;
    open($handle, $cesta) or die("Cannot read file $cesta: $!\n");
    return $handle;
}



#------------------------------------------------------------------------------
# Otevře soubor na disku ke psaní. Jestliže jméno souboru končí na .gz, funkce
# ho při psaní zagzipuje. Vrátí handle souboru. Hodí výjimku, jestliže soubor
# nejde otevřít (chybná cesta, nedostatečná práva atd.)
#------------------------------------------------------------------------------
sub gwopen
{
    my $cesta = shift;
    $cesta =~ s/^\s*>\s*//;
    if($cesta =~ m/\.gz$/)
    {
        $cesta = "| gzip -c > $cesta";
    }
    else
    {
        $cesta = ">$cesta";
    }
    # Jestliže open() dostane místo handlu neinicializovanou lokální skalární
    # proměnnou, vyrobí a vloží do ní anonymní lokální handle, který sám zanikne,
    # až skončí aktuální blok a zaniknou všechny odkazy na něj. (Pokud tedy
    # volající nechce, aby mu handle zanikl, musí ho vložit do globální proměnné.)
    my $handle;
    open($handle, $cesta) or die("Cannot read file $cesta: $!\n");
    return $handle;
}



#------------------------------------------------------------------------------
# Calls an external program. Uses system(). In addition, echoes the command
# line to the standard error output, and returns true/false according to
# whether the call was successful and the external program returned 0 (success)
# or non-zero (error).
#
# Typically called as follows:
#     saferun($command) or die;
#------------------------------------------------------------------------------
sub saferun
{
    my $command = join(' ', @_);
    print STDERR ("Executing: $command\n");
    system($command);
    # The external program does not exist, is not executable or the execution failed for other reasons.
    if($?==-1)
    {
        die("ERROR: Failed to execute: $command\n  $!\n");
    }
    # We were able to start the external program but its execution failed.
    elsif($? & 127)
    {
        printf STDERR ("ERROR: Execution of: $command\n  died with signal %d, %s coredump\n",
            ($? & 127), ($? & 128) ? 'with' : 'without');
        die;
    }
    # The external program ended "successfully" (this still does not guarantee
    # that the external program returned zero!)
    else
    {
        my $exitcode = $? >> 8;
        print STDERR ("Exit code: $exitcode\n") if($exitcode);
        # Return false if the program returned a non-zero value.
        # It is up to the caller how they will handle the return value.
        # (The easiest is to always write:
        # saferun($command) or die;
        # )
        return ! $exitcode;
    }
}



#------------------------------------------------------------------------------
# A variant of saferun() for cases where we normally want to call `command` and
# read its STDOUT.
#------------------------------------------------------------------------------
sub safeticks
{
    my $command = join(' ', @_);
    my $tmp = "/tmp/safeticks.$$";
    saferun("( $command ) > $tmp") or die;
    my $result = `cat $tmp`;
    unlink($tmp) or die("Cannot remove $tmp: $!\n");
    return $result;
}



1;
