#!/usr/bin/perl
# Input Method Editor
# Vybrané řetězce v textu nahradí znaky nebo jinými řetězci.
# Slouží ke snadnému zadávání znaků z cizích abeced pomocí české klávesnice.
# (c) 2008 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

package ime;
use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");



#------------------------------------------------------------------------------
# Sestaví převodní tabulku pro písmena odvozená z latinky s diakritikou.
#------------------------------------------------------------------------------
sub sestavit_tabulku_latin
{
    my %tabulka =
    (
        # Písmena s vodorovnou čárkou nahoře (makronem).
        '-Amac-' => "\x{100}", # LATIN CAPITAL LETTER A WITH MACRON
        '-amac-' => "\x{101}", # LATIN SMALL LETTER A WITH MACRON
        '-amactil-' => "\x{101}\x{303}", # a s makronem a tildou
        '-Emac-' => "\x{112}", # LATIN CAPITAL LETTER E WITH MACRON
        '-emac-' => "\x{113}", # LATIN SMALL LETTER E WITH MACRON
        '-Imac-' => "\x{12A}", # LATIN CAPITAL LETTER I WITH MACRON
        '-imac-' => "\x{12B}", # LATIN SMALL LETTER I WITH MACRON
        '-Omac-' => "\x{14C}", # LATIN CAPITAL LETTER O WITH MACRON
        '-omac-' => "\x{14D}", # LATIN SMALL LETTER O WITH MACRON
        '-Umac-' => "\x{16A}", # LATIN CAPITAL LETTER U WITH MACRON
        '-umac-' => "\x{16B}", # LATIN SMALL LETTER U WITH MACRON
        # Písmena s čárkou doleva (gravem).
        '-Agra-' => "\x{C0}", # LATIN CAPITAL LETTER A WITH GRAVE
        '-agra-' => "\x{E0}", # LATIN SMALL LETTER A WITH GRAVE
        '-Egra-' => "\x{C8}", # LATIN CAPITAL LETTER E WITH GRAVE
        '-egra-' => "\x{E8}", # LATIN SMALL LETTER E WITH GRAVE
        '-Igra-' => "\x{CC}", # LATIN CAPITAL LETTER I WITH GRAVE
        '-igra-' => "\x{EC}", # LATIN SMALL LETTER I WITH GRAVE
        '-Ogra-' => "\x{D2}", # LATIN CAPITAL LETTER O WITH GRAVE
        '-ogra-' => "\x{F2}", # LATIN SMALL LETTER O WITH GRAVE
        '-Ugra-' => "\x{D9}", # LATIN CAPITAL LETTER U WITH GRAVE
        '-ugra-' => "\x{F9}", # LATIN SMALL LETTER U WITH GRAVE
        # Písmena se stříškou (circumflex).
        '-Acir-' => "\x{C2}", # LATIN CAPITAL LETTER A WITH CIRCUMFLEX
        '-acir-' => "\x{E2}", # LATIN SMALL LETTER A WITH CIRCUMFLEX
        '-Ecir-' => "\x{CA}", # LATIN CAPITAL LETTER E WITH CIRCUMFLEX
        '-ecir-' => "\x{EA}", # LATIN SMALL LETTER E WITH CIRCUMFLEX
        '-Icir-' => "\x{CE}", # LATIN CAPITAL LETTER I WITH CIRCUMFLEX
        '-icir-' => "\x{EE}", # LATIN SMALL LETTER I WITH CIRCUMFLEX
        '-Ocir-' => "\x{D4}", # LATIN CAPITAL LETTER O WITH CIRCUMFLEX
        '-ocir-' => "\x{F4}", # LATIN SMALL LETTER O WITH CIRCUMFLEX
        '-Ucir-' => "\x{DB}", # LATIN CAPITAL LETTER U WITH CIRCUMFLEX
        '-ucir-' => "\x{FB}", # LATIN SMALL LETTER U WITH CIRCUMFLEX
        # Písmena s přehláskou (umlaut, dieresis).
        '-I(di[ae]|uml)-' => "\x{CF}", # LATIN CAPITAL LETTER I WITH DIAERESIS
        '-i(di[ae]|uml)-' => "\x{EF}", # LATIN SMALL LETTER I WITH DIAERESIS
        # Písmena s vlnovkou (tildou).
        '-Atil-' => "\x{C3}", # LATIN CAPITAL LETTER A WITH TILDE
        '-atil-' => "\x{E3}", # LATIN SMALL LETTER A WITH TILDE
        '-Itil-' => "\x{128}", # LATIN CAPITAL LETTER I WITH TILDE
        '-itil-' => "\x{129}", # LATIN SMALL LETTER I WITH TILDE
        '-Ntil-' => "\x{D1}", # LATIN CAPITAL LETTER N WITH TILDE
        '-ntil-' => "\x{F1}", # LATIN SMALL LETTER N WITH TILDE
        '-Otil-' => "\x{D5}", # LATIN CAPITAL LETTER O WITH TILDE
        '-otil-' => "\x{F5}", # LATIN SMALL LETTER O WITH TILDE
        '-Util-' => "\x{168}", # LATIN CAPITAL LETTER U WITH TILDE
        '-util-' => "\x{169}", # LATIN SMALL LETTER U WITH TILDE
        # Písmena s ocáskem doleva (cedillou, např. ve francouzštině).
        '-Cced-' => "\x{C7}", # LATIN CAPITAL LETTER C WITH CEDILLA
        '-cced-' => "\x{E7}", # LATIN SMALL LETTER C WITH CEDILLA
        '-Dced-' => "D\x{327}", # LATIN CAPITAL LETTER D + COMBINING CEDILLA
        '-dced-' => "d\x{327}", # LATIN SMALL LETTER D + COMBINING CEDILLA
        '-Gced-' => "\x{122}", # LATIN CAPITAL LETTER G WITH CEDILLA
        '-gced-' => "\x{123}", # LATIN SMALL LETTER G WITH CEDILLA
        '-Hced-' => "H\x{327}", # LATIN CAPITAL LETTER H + COMBINING CEDILLA
        '-hced-' => "h\x{327}", # LATIN SMALL LETTER H + COMBINING CEDILLA
        '-Kced-' => "\x{136}", # LATIN CAPITAL LETTER K WITH CEDILLA
        '-kced-' => "\x{137}", # LATIN SMALL LETTER K WITH CEDILLA
        '-Lced-' => "\x{13B}", # LATIN CAPITAL LETTER L WITH CEDILLA
        '-lced-' => "\x{13C}", # LATIN SMALL LETTER L WITH CEDILLA
        '-Nced-' => "\x{145}", # LATIN CAPITAL LETTER N WITH CEDILLA
        '-nced-' => "\x{146}", # LATIN SMALL LETTER N WITH CEDILLA
        '-Rced-' => "\x{156}", # LATIN CAPITAL LETTER R WITH CEDILLA
        '-rced-' => "\x{157}", # LATIN SMALL LETTER R WITH CEDILLA
        '-Sced-' => "\x{15E}", # LATIN CAPITAL LETTER S WITH CEDILLA
        '-sced-' => "\x{15F}", # LATIN SMALL LETTER S WITH CEDILLA
        '-Tced-' => "\x{162}", # LATIN CAPITAL LETTER T WITH CEDILLA
        '-tced-' => "\x{163}", # LATIN SMALL LETTER T WITH CEDILLA
        # Písmena s oblým háčkem (breve, např. rumunské a nebo turecké g).
        '-Abre-' => "\x{102}", # LATIN CAPITAL LETTER A WITH BREVE
        '-abre-' => "\x{103}", # LATIN SMALL LETTER A WITH BREVE
        '-Gbre-' => "\x{11E}", # LATIN CAPITAL LETTER G WITH BREVE
        '-gbre-' => "\x{11F}", # LATIN SMALL LETTER G WITH BREVE
        # Písmena s tečkou pod (např. pro transliteraci).
        '-ddob-' => "d\x{323}", # d s tečkou pod
        '-hdob-' => "h\x{323}", # h s tečkou pod
        '-ldob-' => "l\x{323}", # l s tečkou pod
        '-mdob-' => "m\x{323}", # m s tečkou pod
        '-ndob-' => "n\x{323}", # n s tečkou pod
        '-rdob-' => "r\x{323}", # r s tečkou pod
        '-sdob-' => "s\x{323}", # s s tečkou pod
        '-tdob-' => "t\x{323}", # t s tečkou pod
        # Písmena s tečkou nad (např. pro transliteraci).
        '-mdot-' => "m\x{307}", # m s tečkou nad
        '-ndot-' => "n\x{307}", # n s tečkou nad
        # Písmena s kroužkem pod (např. r pro transliteraci z dévanágarí).
        '-rrib-' => "r\x{325}", # r s kroužkem pod
        # Další severo- a západoevropská písmena.
        '-Arin-' => "\x{C5}", # LATIN CAPITAL LETTER WITH RING ABOVE
        '-arin-' => "\x{E5}", # LATIN SMALL LETTER WITH RING ABOVE
        '-AE-' => "\x{C6}", # LATIN CAPITAL LIGATURE AE
        '-ae-' => "\x{E6}", # LATIN SMALL LIGATURE AE
        '-Disl-' => "\x{D0}", # LATIN CAPITAL LETTER ETH
        '-disl-' => "\x{F0}", # LATIN SMALL LETTER ETH
        '-Idot-' => "\x{130}", # LATIN CAPITAL LETTER I WITH DOT ABOVE
        '-idot-' => "\x{131}", # LATIN SMALL LETTER DOTLESS I
        '-Osla-' => "\x{D8}", # LATIN CAPITAL LETTER O WITH STROKE
        '-osla-' => "\x{F8}", # LATIN SMALL LETTER O WITH STROKE
        '-OE-' => "\x{152}", # LATIN CAPITAL LIGATURE OE
        '-oe-' => "\x{153}", # LATIN SMALL LIGATURE OE
        '-TH-' => "\x{DE}", # LATIN CAPITAL LETTER THORN
        '-th-' => "\x{FE}", # LATIN SMALL LETTER THORN
        '-ss-' => "\x{DF}", # LATIN SMALL LETTER SHARP S
    );
    return \%tabulka;
}



#------------------------------------------------------------------------------
# Sestaví převodní tabulku z latinky do arabského písma.
#------------------------------------------------------------------------------
sub sestavit_tabulku_arab
{
    # Skupiny znaků, které v arabštině označují podobné hlásky a v češtině je obtížné je rozlišit:
    # t = teh,  T = tah
    # d = dal,  D = dad
    # s = seen, S = sad
    # z = zain, Z = zah
    # h = heh,  H = hah
    # Krátké a, i, u budeme přepisovat jako prázdný řetězec, protože krátké samohlásky se v arabštině obvykle vynechávají (i když na ně jsou prostředky).
    # Následující tabulka přiřazuje každému řetězci v latince kód nebo posloupnost kódů (decimální Unicode) odpovídajících arabských znaků.
    my @arab =
    (
        "a" => "",
        "e" => "",
        "i" => "",
        "u" => "",
        "'" => 1569, # hamza
        "á" => 1575, # alef
        "Á" => 1570, # alef madda
        "'á"=> 1571, # alef hamza above
        "Í" => 1573, # alef hamza below
        "A" => 1614, # fatha (krátké a)
        "U" => 1615, # damma (krátké u)
        "I" => 1616, # kasra (krátké i)
        "__"=> 1617, # shadda (zdvojená souhláska)
        "_" => 1618, # sukun (žádná samohláska)
        "_á_" => 1648, # superscript alef (neznám užití)
        "b" => 1576, # beh
        "ah"=> 1577, # teh marbuta
        "t" => 1578, # teh
        "th"=> 1579, # theh
        "j" => 1580, # jeem
        "dž"=> 1580,
        "H" => 1581, # hah
        "kh"=> 1582, # khah
        "d" => 1583, # dal
        "dh"=> 1584, # thal
        "r" => 1585, # reh
        "z" => 1586, # zain
        "s" => 1587, # seen
        "sh"=> 1588, # sheen
        "š" => 1588,
        "S" => 1589, # sad
        "D" => 1590, # dad
        "T" => 1591, # tah
        "Z" => 1592, # zah
        "c" => 1593, # ain
        "gh"=> 1594, # ghain
        "f" => 1601, # feh
        "q" => 1602, # qaf
        "k" => 1603, # kaf
        "g" => 1709, # berberské g, jak jsem ho viděl v Maroku; podle Unikódu je tohle (3 tečky nahoře) ujgurské ng a to g má mít tečky dole (kód 1710)
        "l" => 1604, # lam
        "m" => 1605, # meem
        "n" => 1606, # noon
        "h" => 1607, # heh
        "w" => 1608, # waw
        "v" => 1608,
        "ú" => 1608,
        "o" => 1608,
        "ó" => 1608,
        "'ú"=> 1572, # waw hamza above
        "ý" => 1609, # alef maksura
        "y" => 1610, # yeh
        "í" => 1610,
        "é" => 1610,
        "'y"=> 1574, # yeh hamza above
    );
    my %tabulka_arab;
    pridat_smisene_pole_kodu(\%tabulka_arab, @arab);
    return \%tabulka_arab;
}



#------------------------------------------------------------------------------
# Sestaví převodní tabulku z latinky do urdsko-arabského písma.
#------------------------------------------------------------------------------
sub sestavit_tabulku_urdu
{
    # Skupiny znaků, které v urdštině označují podobné hlásky a v češtině je obtížné je rozlišit:
    # t = teh, Ť = tah, T = tteh, ť = theh
    # d = dal, Ď = dad, D = ddal, ď = thal
    # h (kh, gh, čh, džh, Th, Dh, th, dh, ph, bh) = heh dvačašmí, h = heh goal, H = hah, x/X = khah
    # é = yeh barrí, í/e = farsí yeh, y/j = yeh hamza
    # n = nún, m = mím, N/M = nún ghunna
    # á = alef, Á = alef madda
    # s = sín, S = sad; z = zajn, Z = zah; r = reh, R = rreh
    # Naopak znak waw lze do latinky přepsat různými způsoby, budeme reagovat na tyto protějšky: v, w, ú, o, ó.
    # Krátké a, i, u budeme přepisovat jako prázdný řetězec, protože krátké samohlásky se v urdštině obvykle vynechávají (i když na ně jsou prostředky).
    # Následující tabulka přiřazuje každému řetězci v latince kód nebo posloupnost kódů (decimální Unicode) odpovídajících urdských znaků.
    my @urdu =
    (
        "a" => "",
        "i" => "",
        "u" => "",
        "á" => 1575, # alef
        "Á" => 1570, # alef madda
        "e" => 1740, # farsi yeh
        "í" => 1740,
        "y" => 1574, # hamza yeh
        "j" => 1574,
        "é" => 1746, # yeh barree
        "v" => 1608, # waw
        "w" => 1608,
        "ú" => 1608,
        "o" => 1608,
        "ó" => 1608,
        "A" => 1614, # fatha (krátké a)
        "U" => 1615, # damma (krátké u)
        "I" => 1616, # kasra (krátké i)
        "_" => 1617, # shadda (žádná samohláska)
        "Í" => 1648, # superscript alef (neznám užití)
        "k" => 1705, # keheh
        "r" => 1585, # reh
        "R" => 1681, # rreh
        "kh" => "1705+1726", # keheh + heh doachashmee
        "gh" => "1711+1726",
        "čh" => "1670+1726",
        "džh" => "1580+1726",
        "Th" => "1657+1726",
        "Dh" => "1672+1726",
        "th" => "1578+1726",
        "dh" => "1583+1726",
        "ph" => "1662+1726",
        "bh" => "1576+1726",
        "h" => 1729, # heh goal
        "H" => 1581, # hah
        "x" => 1582, # khah
        "X" => 1582,
        "n" => 1606, # noon
        "m" => 1605, # meem
        "N" => 1722, # noon ghunna
        "M" => 1722,
        "t" => 1578, # teh
        "T" => 1657, # tteh
        "Ť" => 1591, # tah
        "ť" => 1579, # theh
        "d" => 1583, # dal
        "D" => 1672, # ddal
        "Ď" => 1590, # dad
        "ď" => 1584, # thal
        "s" => 1587, # seen
        "S" => 1589, # sad
        "l" => 1604, # lam
        "b" => 1576, # beh
        "dž" => 1580, # jeem
        "p" => 1662, # peh
        "." => 1748, # full stop
        "," => 1548, # comma
        "?" => 1567, # question
        "g" => 1711, # gaf
        "c" => 1593, # ain
        "C" => 1593,
        "q" => 1602, # qaf
        "š" => 1588, # sheen
        "f" => 1601, # feh
        "z" => 1586, # zain
        "Z" => 1592, # zah
        "č" => 1670, # tcheh
        "G" => 1594, # ghain
    );
    my %tabulka_urdu;
    pridat_smisene_pole_kodu(\%tabulka_urdu, @urdu);
    return \%tabulka_urdu;
}



#------------------------------------------------------------------------------
# Sestaví převodní tabulku z latinky do dévanágarí.
#------------------------------------------------------------------------------
sub sestavit_tabulku_devanagari
{
    my %tabulka_devanagari;
    # Přidat do transliterační tabulky samostatné samohlásky.
    my @samohlasky = ("a", "á", "i", "í", "u", "ú", "ŕ", "ĺ", "E", "e", "é", "aj", "O", "o", "ó", "au");
    pridat_zdrojove_pole(\%tabulka_devanagari, hex("905"), @samohlasky);
    my @souhlasky =
    (
        "k", "kh", "g",  "gh",  "ng",
        "č", "čh", "dž", "džh", "ň",
        "T", "Th", "D",  "Dh",  "N",
        "t", "th", "d",  "dh",  "n", "ń",
        "p", "ph", "b",  "bh",  "m",
        "j", "r",  "RR",  "l",   "L", "ł", "v",
        "ś", "š",  "s",  "h"
    );
    my @souhlasky2 = ("q", "Kh", "Gh", "z", "R", "Rh", "f", "Y");
    pridat_zdrojove_pole(\%tabulka_devanagari, hex("915"), map {$_."a"} @souhlasky);
    pridat_zdrojove_pole(\%tabulka_devanagari, hex("958"), map {$_."a"} @souhlasky2);
    pridat_dve_pole(\%tabulka_devanagari, (map {$_."a"} @souhlasky), (map {chr($_)} (hex("915") .. hex("939"))));
    pridat_dve_pole(\%tabulka_devanagari, (map {$_."a"} @souhlasky2), (map {chr($_)} (hex("958") .. hex("95F"))));
    my @diahlasky = ("á", "i", "í", "u", "ú", "ŕ", "Ŕ", "E", "e", "é", "aj", "O", "o", "ó", "au", ""); # poslední je virám
    for(my $i = 0; $i<=$#diahlasky; $i++)
    {
        my $dia0 = $diahlasky[$i];
        my $dia1 = chr(hex("93E")+$i);
        pridat_dve_pole(\%tabulka_devanagari, (map {$_.$dia0} @souhlasky), (map {chr($_).$dia1} (hex("915") .. hex("939"))));
        pridat_dve_pole(\%tabulka_devanagari, (map {$_.$dia0} @souhlasky2), (map {chr($_).$dia1} (hex("958") .. hex("95F"))));
    }
    # Přidat anusvár (bindu).
    $tabulka_devanagari{"M"} = chr(hex("902"));
    # Přidat anunásik (čandrabindu).
    $tabulka_devanagari{"MM"} = chr(hex("901"));
    # Přidat visarg (přídech).
    $tabulka_devanagari{"H"} = chr(hex("903"));
    # Přidat číslice.
    pridat_zdrojove_pole(\%tabulka_devanagari, hex("966"), (0..9));
    # Přidat do transliterační tabulky dandu (značka dévanágarí místo evropské tečky za větou).
    $tabulka_devanagari{"."} = chr(hex("964"));
    return \%tabulka_devanagari;
}



#------------------------------------------------------------------------------
# Sestaví převodní tabulku z latinky do gurmukhí.
#------------------------------------------------------------------------------
sub sestavit_tabulku_gurmukhi
{
    my %tabulka_gurmukhi;
    # Přidat do transliterační tabulky samostatné samohlásky.
    my @samohlasky = ("a", "á", "i", "í", "u", "ú", "ŕ", "ĺ", "E", "e", "é", "aj", "O", "o", "ó", "au");
    pridat_zdrojove_pole(\%tabulka_gurmukhi, hex("A05"), @samohlasky);
    my @souhlasky =
    (
        "k", "kh", "g",  "gh",  "ng",
        "č", "čh", "dž", "džh", "ň",
        "T", "Th", "D",  "Dh",  "N",
        "t", "th", "d",  "dh",  "n", "ń",
        "p", "ph", "b",  "bh",  "m",
        "j", "r",  "RR",  "l",   "L", "ł", "v",
        "ś", "š",  "s",  "h"
    );
    my @souhlasky2 = ("q", "Kh", "Gh", "z", "R", "Rh", "f", "Y");
    pridat_zdrojove_pole(\%tabulka_gurmukhi, hex("A15"), map {$_."a"} @souhlasky);
    pridat_zdrojove_pole(\%tabulka_gurmukhi, hex("A58"), map {$_."a"} @souhlasky2);
    pridat_dve_pole(\%tabulka_gurmukhi, (map {$_."a"} @souhlasky), (map {chr($_)} (hex("A15") .. hex("A39"))));
    pridat_dve_pole(\%tabulka_gurmukhi, (map {$_."a"} @souhlasky2), (map {chr($_)} (hex("A58") .. hex("A5F"))));
    my @diahlasky = ("á", "i", "í", "u", "ú", "ŕ", "Ŕ", "E", "e", "é", "aj", "O", "o", "ó", "au", ""); # poslední je virám
    for(my $i = 0; $i<=$#diahlasky; $i++)
    {
        my $dia0 = $diahlasky[$i];
        my $dia1 = chr(hex("A3E")+$i);
        pridat_dve_pole(\%tabulka_gurmukhi, (map {$_.$dia0} @souhlasky), (map {chr($_).$dia1} (hex("A15") .. hex("A39"))));
        pridat_dve_pole(\%tabulka_gurmukhi, (map {$_.$dia0} @souhlasky2), (map {chr($_).$dia1} (hex("A58") .. hex("A5F"))));
    }
    # Přidat anusvár (bindu).
    $tabulka_gurmukhi{"M"} = chr(hex("A02"));
    # Přidat anunásik (čandrabindu).
    $tabulka_gurmukhi{"MM"} = chr(hex("A01"));
    # Přidat visarg (přídech).
    $tabulka_gurmukhi{"H"} = chr(hex("A03"));
    # Přidat číslice.
    pridat_zdrojove_pole(\%tabulka_gurmukhi, hex("A66"), (0..9));
    # Přidat do transliterační tabulky dandu (značka dévanágarí místo evropské tečky za větou).
    $tabulka_gurmukhi{"."} = chr(hex("A64"));
    return \%tabulka_gurmukhi;
}



#------------------------------------------------------------------------------
# Sestaví převodní tabulku z latinky do telugu.
#------------------------------------------------------------------------------
sub sestavit_tabulku_telugu
{
    my $pocatek_samohlasek = 3077; # a
    my $pocatek_souhlasek = 3093; # ka
    my $konec_souhlasek = 3129; # ha
    my $pocatek_souhlasek_2 = 3160; # v telugštině se nepoužívají, v dévanágarí ano
    my $konec_souhlasek_2 = 3167;
    my $pocatek_diahlasek = 3134;
    my $anusvar = 3074;
    my $pocatek_cislic = 3174;
    my %tabulka_telugu;
    # Přidat do transliterační tabulky samostatné samohlásky.
    my @samohlasky = ("a", "á", "i", "í", "u", "ú", "ŕ", "ĺ", "E", "e", "é", "aj", "O", "o", "ó", "au");
    pridat_zdrojove_pole(\%tabulka_telugu, $pocatek_samohlasek, @samohlasky);
    # Přidat do transliterační tabulky souhlásky s implicitním "a".
    my @souhlasky =
    (
        "k", "kh", "g",  "gh",  "ng",
        "č", "čh", "dž", "džh", "ň",
        "T", "Th", "D",  "Dh",  "N",
        "t", "th", "d",  "dh",  "n", "ń",
        "p", "ph", "b",  "bh",  "m",
        "j", "r",  "R",  "l",   "L", "ł", "v",
        "ś", "š",  "s",  "h"
    );
    my @souhlasky2 = ("q", "Kh", "Gh", "z", "DDDH", "RH", "f", "Y");
    pridat_zdrojove_pole(\%tabulka_telugu, $pocatek_souhlasek, map {$_."a"} @souhlasky);
    pridat_zdrojove_pole(\%tabulka_telugu, $pocatek_souhlasek_2, map {$_."a"} @souhlasky2);
    # Přidat do transliterační tabulky souhlásky s ostatními samohláskami a s virámem.
    my @diahlasky = ("á", "i", "í", "u", "ú", "ŕ", "Ŕ", "E", "e", "é", "aj", "O", "o", "ó", "au", ""); # poslední je virám
    for(my $i = 0; $i<=$#diahlasky; $i++)
    {
        my $dia0 = $diahlasky[$i];
        my $dia1 = chr($pocatek_diahlasek+$i);
        pridat_dve_pole(\%tabulka_telugu, (map {$_.$dia0} @souhlasky), (map {chr($_).$dia1} ($pocatek_souhlasek .. $konec_souhlasek)));
        pridat_dve_pole(\%tabulka_telugu, (map {$_.$dia0} @souhlasky2), (map {chr($_).$dia1} ($pocatek_souhlasek_2 .. $konec_souhlasek_2)));
    }
    # Přidat do transliterační tabulky anusvár.
    $tabulka_telugu{"M"} = chr($anusvar);
    # Přidat do transliterační tabulky dandu (značka dévanágarí místo evropské tečky za větou).
    # Problém: v telugštině se používá normální tečka, takže nechceme, aby se nám tečka přepsala na neexistující znak na pozici odpovídající dandě!
    # Přidat do transliterační tabulky číslice.
    pridat_zdrojove_pole(\%tabulka_telugu, $pocatek_cislic, '0', '1', '2', '3', '4', '5', '6', '7', '8', '9');
    return \%tabulka_telugu;
}



#------------------------------------------------------------------------------
# Ve vstupním textu přepíše pomocné sekvence cílovými znaky a řetězec vrátí.
#------------------------------------------------------------------------------
sub prepsat
{
    my $text = shift;
    # Písmena s ostrým (českým) háčkem. (I ta česká se mohou hodit lidem, kteří nemají českou klávesnici.)
    $text =~ s/-Ccar-/\x{10C}/g; # LATIN CAPITAL LETTER C WITH CARON
    $text =~ s/-ccar-/\x{10D}/g; # LATIN SMALL LETTER C WITH CARON
    $text =~ s/-Scar-/\x{160}/g; # LATIN CAPITAL LETTER S WITH CARON
    $text =~ s/-scar-/\x{161}/g; # LATIN SMALL LETTER S WITH CARON
    $text =~ s/-Zcar-/\x{17D}/g; # LATIN CAPITAL LETTER Z WITH CARON
    $text =~ s/-zcar-/\x{17E}/g; # LATIN SMALL LETTER Z WITH CARON
    # Písmena s vodorovnou čárkou nahoře (makronem).
    $text =~ s/-Amac-/\x{100}/g; # LATIN CAPITAL LETTER A WITH MACRON
    $text =~ s/-amac-/\x{101}/g; # LATIN SMALL LETTER A WITH MACRON
    $text =~ s/-amactil-/\x{101}\x{303}/g; # a s makronem a tildou
    $text =~ s/-Emac-/\x{112}/g; # LATIN CAPITAL LETTER E WITH MACRON
    $text =~ s/-emac-/\x{113}/g; # LATIN SMALL LETTER E WITH MACRON
    $text =~ s/-emactil-/\x{113}\x{303}/g; # e s makronem a tildou
    $text =~ s/-Imac-/\x{12A}/g; # LATIN CAPITAL LETTER I WITH MACRON
    $text =~ s/-imac-/\x{12B}/g; # LATIN SMALL LETTER I WITH MACRON
    $text =~ s/-imactil-/\x{12B}\x{303}/g; # i s makronem a tildou
    $text =~ s/-Omac-/\x{14C}/g; # LATIN CAPITAL LETTER O WITH MACRON
    $text =~ s/-omac-/\x{14D}/g; # LATIN SMALL LETTER O WITH MACRON
    $text =~ s/-omactil-/\x{14D}\x{303}/g; # o s makronem a tildou
    $text =~ s/-Umac-/\x{16A}/g; # LATIN CAPITAL LETTER U WITH MACRON
    $text =~ s/-umac-/\x{16B}/g; # LATIN SMALL LETTER U WITH MACRON
    $text =~ s/-umactil-/\x{16B}\x{303}/g; # u s makronem a tildou
    # Písmena s čárkou doleva (gravem).
    $text =~ s/-Agra-/\x{C0}/g; # LATIN CAPITAL LETTER A WITH GRAVE
    $text =~ s/-agra-/\x{E0}/g; # LATIN SMALL LETTER A WITH GRAVE
    $text =~ s/-Egra-/\x{C8}/g; # LATIN CAPITAL LETTER E WITH GRAVE
    $text =~ s/-egra-/\x{E8}/g; # LATIN SMALL LETTER E WITH GRAVE
    $text =~ s/-Igra-/\x{CC}/g; # LATIN CAPITAL LETTER I WITH GRAVE
    $text =~ s/-igra-/\x{EC}/g; # LATIN SMALL LETTER I WITH GRAVE
    $text =~ s/-Ogra-/\x{D2}/g; # LATIN CAPITAL LETTER O WITH GRAVE
    $text =~ s/-ogra-/\x{F2}/g; # LATIN SMALL LETTER O WITH GRAVE
    $text =~ s/-Ugra-/\x{D9}/g; # LATIN CAPITAL LETTER U WITH GRAVE
    $text =~ s/-ugra-/\x{F9}/g; # LATIN SMALL LETTER U WITH GRAVE
    # Písmena se stříškou (circumflex).
    $text =~ s/-Acir-/\x{C2}/g; # LATIN CAPITAL LETTER A WITH CIRCUMFLEX
    $text =~ s/-acir-/\x{E2}/g; # LATIN SMALL LETTER A WITH CIRCUMFLEX
    $text =~ s/-Ecir-/\x{CA}/g; # LATIN CAPITAL LETTER E WITH CIRCUMFLEX
    $text =~ s/-ecir-/\x{EA}/g; # LATIN SMALL LETTER E WITH CIRCUMFLEX
    $text =~ s/-Icir-/\x{CE}/g; # LATIN CAPITAL LETTER I WITH CIRCUMFLEX
    $text =~ s/-icir-/\x{EE}/g; # LATIN SMALL LETTER I WITH CIRCUMFLEX
    $text =~ s/-Ocir-/\x{D4}/g; # LATIN CAPITAL LETTER O WITH CIRCUMFLEX
    $text =~ s/-ocir-/\x{F4}/g; # LATIN SMALL LETTER O WITH CIRCUMFLEX
    $text =~ s/-Ucir-/\x{DB}/g; # LATIN CAPITAL LETTER U WITH CIRCUMFLEX
    $text =~ s/-ucir-/\x{FB}/g; # LATIN SMALL LETTER U WITH CIRCUMFLEX
    # Písmena s přehláskou (umlaut, dieresis).
    $text =~ s/-I(di[ae]|uml)-/\x{CF}/g; # LATIN CAPITAL LETTER I WITH DIAERESIS
    $text =~ s/-i(di[ae]|uml)-/\x{EF}/g; # LATIN SMALL LETTER I WITH DIAERESIS
    # Písmena s vlnovkou (tildou).
    $text =~ s/-Atil-/\x{C3}/g; # LATIN CAPITAL LETTER A WITH TILDE
    $text =~ s/-atil-/\x{E3}/g; # LATIN SMALL LETTER A WITH TILDE
    $text =~ s/-Etil-/\x{1EBC}/g; # LATIN CAPITAL LETTER E WITH TILDE
    $text =~ s/-etil-/\x{1EBD}/g; # LATIN SMALL LETTER E WITH TILDE
    $text =~ s/-Itil-/\x{128}/g; # LATIN CAPITAL LETTER I WITH TILDE
    $text =~ s/-itil-/\x{129}/g; # LATIN SMALL LETTER I WITH TILDE
    $text =~ s/-Ntil-/\x{D1}/g; # LATIN CAPITAL LETTER N WITH TILDE
    $text =~ s/-ntil-/\x{F1}/g; # LATIN SMALL LETTER N WITH TILDE
    $text =~ s/-Otil-/\x{D5}/g; # LATIN CAPITAL LETTER O WITH TILDE
    $text =~ s/-otil-/\x{F5}/g; # LATIN SMALL LETTER O WITH TILDE
    $text =~ s/-Util-/\x{168}/g; # LATIN CAPITAL LETTER U WITH TILDE
    $text =~ s/-util-/\x{169}/g; # LATIN SMALL LETTER U WITH TILDE
    # Písmena s ocáskem doleva (cedillou, např. ve francouzštině).
    $text =~ s/-Cced-/\x{C7}/g; # LATIN CAPITAL LETTER C WITH CEDILLA
    $text =~ s/-cced-/\x{E7}/g; # LATIN SMALL LETTER C WITH CEDILLA
    $text =~ s/-Dced-/D\x{327}/g; # LATIN CAPITAL LETTER D + COMBINING CEDILLA
    $text =~ s/-dced-/d\x{327}/g; # LATIN SMALL LETTER D + COMBINING CEDILLA
    $text =~ s/-Gced-/\x{122}/g; # LATIN CAPITAL LETTER G WITH CEDILLA
    $text =~ s/-gced-/\x{123}/g; # LATIN SMALL LETTER G WITH CEDILLA
    $text =~ s/-Hced-/H\x{327}/g; # LATIN CAPITAL LETTER H + COMBINING CEDILLA
    $text =~ s/-hced-/h\x{327}/g; # LATIN SMALL LETTER H + COMBINING CEDILLA
    $text =~ s/-Kced-/\x{136}/g; # LATIN CAPITAL LETTER K WITH CEDILLA
    $text =~ s/-kced-/\x{137}/g; # LATIN SMALL LETTER K WITH CEDILLA
    $text =~ s/-Lced-/\x{13B}/g; # LATIN CAPITAL LETTER L WITH CEDILLA
    $text =~ s/-lced-/\x{13C}/g; # LATIN SMALL LETTER L WITH CEDILLA
    $text =~ s/-Nced-/\x{145}/g; # LATIN CAPITAL LETTER N WITH CEDILLA
    $text =~ s/-nced-/\x{146}/g; # LATIN SMALL LETTER N WITH CEDILLA
    $text =~ s/-Rced-/\x{156}/g; # LATIN CAPITAL LETTER R WITH CEDILLA
    $text =~ s/-rced-/\x{157}/g; # LATIN SMALL LETTER R WITH CEDILLA
    $text =~ s/-Sced-/\x{15E}/g; # LATIN CAPITAL LETTER S WITH CEDILLA
    $text =~ s/-sced-/\x{15F}/g; # LATIN SMALL LETTER S WITH CEDILLA
    $text =~ s/-Tced-/\x{162}/g; # LATIN CAPITAL LETTER T WITH CEDILLA
    $text =~ s/-tced-/\x{163}/g; # LATIN SMALL LETTER T WITH CEDILLA
    $text =~ s/-Zced-/Z\x{327}/g; # LATIN CAPITAL LETTER Z + COMBINING CEDILLA
    $text =~ s/-zced-/z\x{327}/g; # LATIN SMALL LETTER Z + COMBINING CEDILLA
    # Písmena s oblým háčkem (breve, např. rumunské a nebo turecké g).
    $text =~ s/-Abre-/\x{102}/g; # LATIN CAPITAL LETTER A WITH BREVE
    $text =~ s/-abre-/\x{103}/g; # LATIN SMALL LETTER A WITH BREVE
    $text =~ s/-Gbre-/\x{11E}/g; # LATIN CAPITAL LETTER G WITH BREVE
    $text =~ s/-gbre-/\x{11F}/g; # LATIN SMALL LETTER G WITH BREVE
    # Písmena s tečkou pod (např. pro transliteraci).
    $text =~ s/-ddob-/\x{1E0D}/g; # LATIN SMALL LETTER D WITH DOT BELOW
    $text =~ s/-hdob-/\x{1E25}/g; # LATIN SMALL LETTER H WITH DOT BELOW
    $text =~ s/-ldob-/\x{1E37}/g; # LATIN SMALL LETTER L WITH DOT BELOW
    $text =~ s/-mdob-/\x{1E43}/g; # LATIN SMALL LETTER M WITH DOT BELOW
    $text =~ s/-ndob-/\x{1E47}/g; # LATIN SMALL LETTER N WITH DOT BELOW
    $text =~ s/-rdob-/\x{1E5B}/g; # LATIN SMALL LETTER R WITH DOT BELOW
    $text =~ s/-sdob-/\x{1E63}/g; # LATIN SMALL LETTER S WITH DOT BELOW
    $text =~ s/-tdob-/\x{1E6D}/g; # LATIN SMALL LETTER T WITH DOT BELOW
    # Písmena s tečkou nad (např. pro transliteraci).
    $text =~ s/-mdot-/\x{1E41}/g; # LATIN SMALL LETTER M WITH DOT ABOVE
    $text =~ s/-ndot-/\x{1E45}/g; # LATIN SMALL LETTER N WITH DOT ABOVE
    $text =~ s/-sdot-/\x{1E61}/g; # LATIN SMALL LETTER S WITH DOT ABOVE
    $text =~ s/-zdot-/\x{17C}/g;  # LATIN SMALL LETTER Z WITH DOT ABOVE
    # Písmena s kroužkem pod (např. r pro transliteraci z dévanágarí).
    $text =~ s/-rrib-/r\x{325}/g; # r s kroužkem pod
    # Další severo- a západoevropská písmena.
    $text =~ s/-Arin-/\x{C5}/g; # LATIN CAPITAL LETTER WITH RING ABOVE
    $text =~ s/-arin-/\x{E5}/g; # LATIN SMALL LETTER WITH RING ABOVE
    $text =~ s/-AE-/\x{C6}/g; # LATIN CAPITAL LIGATURE AE
    $text =~ s/-ae-/\x{E6}/g; # LATIN SMALL LIGATURE AE
    $text =~ s/-Disl-/\x{D0}/g; # LATIN CAPITAL LETTER ETH
    $text =~ s/-disl-/\x{F0}/g; # LATIN SMALL LETTER ETH
    $text =~ s/-Idot-/\x{130}/g; # LATIN CAPITAL LETTER I WITH DOT ABOVE
    $text =~ s/-idot-/\x{131}/g; # LATIN SMALL LETTER DOTLESS I
    $text =~ s/-Osla-/\x{D8}/g; # LATIN CAPITAL LETTER O WITH STROKE
    $text =~ s/-osla-/\x{F8}/g; # LATIN SMALL LETTER O WITH STROKE
    $text =~ s/-OE-/\x{152}/g; # LATIN CAPITAL LIGATURE OE
    $text =~ s/-oe-/\x{153}/g; # LATIN SMALL LIGATURE OE
    $text =~ s/-TH-/\x{DE}/g; # LATIN CAPITAL LETTER THORN
    $text =~ s/-th-/\x{FE}/g; # LATIN SMALL LETTER THORN
    $text =~ s/-ss-/\x{DF}/g; # LATIN SMALL LETTER SHARP S
    # Další písmena.
    $text =~ s/-glot-/\x{2C0}/g; # MODIFIER LETTER GLOTTAL STOP # používám pro transliteraci arabského ajnu
    # Přepis do cyrilice.
    while($text =~ m/-(rus|ukr)1-(.*?)-(?:\1)0-/s)
    {
        my $jazyk = $1;
        my $latinka = $2;
        my $cyrilice;
        while($latinka)
        {
            if($latinka =~ s/^(ŠČ|Šč)//)
            {
                $cyrilice .= "\x{429}";
            }
            elsif($latinka =~ s/^šč//)
            {
                $cyrilice .= "\x{449}";
            }
            elsif($latinka =~ s/^(JU|Ju)//)
            {
                $cyrilice .= "\x{42E}";
            }
            elsif($latinka =~ s/^ju//)
            {
                $cyrilice .= "\x{44E}";
            }
            elsif($latinka =~ s/^(JA|Ja)//)
            {
                $cyrilice .= "\x{42F}";
            }
            elsif($latinka =~ s/^ja//)
            {
                $cyrilice .= "\x{44F}";
            }
            elsif($latinka =~ s/^(JE|Je)//)
            {
                $cyrilice .= "\x{404}";
            }
            elsif($latinka =~ s/^je//)
            {
                $cyrilice .= "\x{454}";
            }
            elsif($latinka =~ s/^(JI|Ji)//)
            {
                $cyrilice .= "\x{407}";
            }
            elsif($latinka =~ s/^ji//)
            {
                $cyrilice .= "\x{457}";
            }
            else
            {
                $latinka =~ s/^(.)//;
                my $znak = $1;
                # Měkký znak: malý ´ (česká čárka nad samohláskou), velký ˇ (český háček)
                # Tvrdý znak: malý ě, velký Ě
                # Tvrdé E: malé é, velké É (přepis pomocí e s čárkou doleva by byl s českou klávesnicí nepohodlný)
                $znak =~ tr/ABVGDEËŽZIJKLMNOPRSTUFHCČŠĚYˇÉabvgdeëžzijklmnoprstufhcčšěy´é/\x{410}\x{411}\x{412}\x{413}\x{414}\x{415}\x{401}\x{416}\x{417}\x{418}\x{419}\x{41A}\x{41B}\x{41C}\x{41D}\x{41E}\x{41F}\x{420}\x{421}\x{422}\x{423}\x{424}\x{425}\x{426}\x{427}\x{428}\x{42A}\x{42B}\x{42C}\x{42D}\x{430}\x{431}\x{432}\x{433}\x{434}\x{435}\x{451}\x{436}\x{437}\x{438}\x{439}\x{43A}\x{43B}\x{43C}\x{43D}\x{43E}\x{43F}\x{440}\x{441}\x{442}\x{443}\x{444}\x{445}\x{446}\x{447}\x{448}\x{44A}\x{44B}\x{44C}\x{44D}/;
                # Některá písmena přepisovat jinak v ukrajinštině.
                if($jazyk eq "ukr")
                {
                    # Posunutý přepis "i" a "y", místo tvrdého znaku apostrof - ten lze také zapsat paragrafem.
                    $znak =~ tr/\x{418}\x{42B}\x{438}\x{44B}\x{42A}\x{44A}§/\x{406}\x{418}\x{456}\x{438}'''/;
                }
                $cyrilice .= $znak;
            }
        }
        $text =~ s/-${jazyk}1-.*?-${jazyk}0-/$cyrilice/s;
    }
    # Přepis do arabštiny.
    my $tabulka_arab = sestavit_tabulku_arab();
    while($text =~ m/-(ara?b?)1-(.*?)-(?:\1)0-/s)
    {
        my $jazyk = $1;
        my $latinka = $2;
        my $arab;
        my $arab = prepsat_tabulka($tabulka_arab, $latinka);
        $text =~ s/-${jazyk}1-.*?-${jazyk}0-/$arab/s;
    }
    # Přepis do urdštiny.
    my $tabulka_urdu = sestavit_tabulku_urdu();
    while($text =~ m/-(urd)1-(.*?)-(?:\1)0-/s)
    {
        my $jazyk = $1;
        my $latinka = $2;
        my $urdu;
        my $urdu = prepsat_tabulka($tabulka_urdu, $latinka);
        $text =~ s/-${jazyk}1-.*?-${jazyk}0-/$urdu/s;
    }
    # Přepis do dévanágarí.
    my $tabulka_devanagari = sestavit_tabulku_devanagari();
    while($text =~ m/-(dev)1-(.*?)-(?:\1)0-/s)
    {
        my $jazyk = $1;
        my $latinka = $2;
        my $devanagari = prepsat_tabulka($tabulka_devanagari, $latinka);
        $text =~ s/-${jazyk}1-.*?-${jazyk}0-/$devanagari/s;
    }
    # Přepis do gurmukhí.
    my $tabulka_gurmukhi = sestavit_tabulku_gurmukhi();
    while($text =~ m/-(gur)1-(.*?)-(?:\1)0-/s)
    {
        my $jazyk = $1;
        my $latinka = $2;
        my $gurmukhi = prepsat_tabulka($tabulka_gurmukhi, $latinka);
        $text =~ s/-${jazyk}1-.*?-${jazyk}0-/$gurmukhi/s;
    }
    # Přepis do telugštiny.
    my $tabulka_telugu = sestavit_tabulku_telugu();
    while($text =~ m/-(tel)1-(.*?)-(?:\1)0-/s)
    {
        my $jazyk = $1;
        my $latinka = $2;
        my $telugu = prepsat_tabulka($tabulka_telugu, $latinka);
        $text =~ s/-${jazyk}1-.*?-${jazyk}0-/$telugu/s;
    }
    # Přepis nelatin1 znaků do entit HTML.
    while($text =~ m/-ent1-(.*?)-ent0-/s)
    {
        my $retezec = $1;
        $retezec = join('', map {$_ = ord($_)>=256 ? '&amp;#'.ord($_).';' : $_} split(//, $retezec));
        $text =~ s/-ent1-.*?-ent0-/$retezec/s;
    }
    # Několik dalších znakových substitucí provést taky tady.
    # Specielně ty ztrojené a zdvojené pomlčky musí být úplně na konci, aby nezablokovaly sousedící jiné entity.
    $text =~ s/-xx-/\x{D7}/g; # MULTIPLICATION SIGN
    $text =~ s/&sup2;/\xB2/g; # SUPERSCRIPT TWO
    $text =~ s/&sup3;/\xB3/g; # SUPERSCRIPT THREE
    $text =~ s/-nb-/&nbsp;/g; # mezera nerozdělující slovo, jako jediná má zůstat jako entita
    $text =~ s/-br-/<br\/>/g;   # pevné zalomení řádku
    # Náhrada spojovníku pomlčkou nebo dlouhou pomlčkou. Dříve jsem používal zkratky "--" a "---".
    # Musely se nahrazovat až jako poslední (em, pak en), aby jejich nahrazení nepoškodilo jiné entity.
    # I tak to kolidovalo se syntaxí HTML a MediaWiki. Teď, když náhrady přesunuji do samostatné
    # knihovny, která má být obecná a o HTML ani MediaWiki raději nic nevědět, raději volím bezpečnější
    # zkratky, než abych zde komplikoval regulární výrazy ve snaze zachránit syntax všeho možného.
    $text =~ s/-nd-/\x{2013}/g; # EN DASH
    $text =~ s/-md-/\x{2014}/g; # EM DASH
#    $text =~ s/---/\x{2014}/g; # EM DASH
    # Nerušit <!-- ... --> (HTML komentáře).
#    $text =~ s/(?<!<!)--(?!>)/\x{2013}/g; # EN DASH
    # Typografické uvozovky.
    $text =~ s/-en"1-/\x{201C}/g; # LEFT DOUBLE QUOTATION MARK
    $text =~ s/-en"0-/\x{201D}/g; # RIGHT DOUBLE QUOTATION MARK
    $text =~ s/-en'1-/\x{2018}/g; # LEFT SINGLE QUOTATION MARK
    $text =~ s/-en'0-/\x{2019}/g; # RIGHT SINGLE QUOTATION MARK
    $text =~ s/-cs"1-/\x{201E}/g; # DOUBLE LOW-9 QUOTATION MARK
    $text =~ s/-cs"0-/\x{201C}/g; # LEFT DOUBLE QUOTATION MARK (who was so dumb to call this left?)
    $text =~ s/-cs'1-/\x{201A}/g; # SINGLE LOW-9 QUOTATION MARK
    $text =~ s/-cs'0-/\x{2018}/g; # LEFT SINGLE QUOTATION MARK (who was so dumb to call this left?)
    $text =~ s/-"-/\x{201C}English\x{201D} \x{201E}česky\x{201C} \x{2018}English\x{2019} \x{201A}česky\x{2018}/g; # the entire repertoire
    return $text;
}



#------------------------------------------------------------------------------
# Převezme pole řetězců, z nichž každý se má přepsat na právě jeden znak
# Unikódu a kódy cílových znaků představují souvislý úsek Unikódu. Dále
# převezme kód prvního cílového znaku a odkaz na hash s převodní tabulkou.
# Tento odkaz také vrátí.
#------------------------------------------------------------------------------
sub pridat_zdrojove_pole
{
    my $tabulka = shift; # odkaz na hash
    my $cil0 = shift; # kód prvního cílového znaku
    my @zdroj = @_; # pole zdrojových řetězců
    for(my $i = 0; $i<=$#zdroj; $i++)
    {
        $tabulka->{$zdroj[$i]} = chr($cil0+$i);
    }
    return $tabulka;
}



#------------------------------------------------------------------------------
# Převezme pole zdrojových řetězců a pole cílových řetězců. Obě pole musí být
# stejně dlouhá, jednotlivé řetězce mohou být libovolně dlouhé. Dále převezme
# odkaz na hash s převodní tabulkou, do které má uložit převody.
#------------------------------------------------------------------------------
sub pridat_dve_pole
{
    my $tabulka = shift; # odkaz na hash
    my $n = ($#_+1)/2;
    my @zdroj = @_[0..($n-1)];
    my @cil = @_[$n..(2*$n-1)];
    for(my $i = 0; $i<$n; $i++)
    {
        $tabulka->{$zdroj[$i]} = $cil[$i];
    }
    return $tabulka;
}



#------------------------------------------------------------------------------
# Převezme pole zdrojových a cílových řetězců. První prvek je zdrojový, druhý
# cílový, třetí zdrojový, čtvrtý cílový atd. Dále převezme odkaz na hash
# s převodní tabulkou, do které má uložit převody.
#------------------------------------------------------------------------------
sub pridat_smisene_pole
{
    my $tabulka = shift; # odkaz na hash
    for(my $i = 0; $i<$#_; $i += 2)
    {
        $tabulka->{$_[$i]} = $_[$i+1];
    }
    return $tabulka;
}



#------------------------------------------------------------------------------
# Převezme pole zdrojových a cílových řetězců. První prvek je zdrojový, druhý
# cílový, třetí zdrojový, čtvrtý cílový atd. Dále převezme odkaz na hash
# s převodní tabulkou, do které má uložit převody.
#------------------------------------------------------------------------------
sub pridat_smisene_pole_kodu
{
    my $tabulka = shift; # odkaz na hash
    for(my $i = 0; $i<$#_; $i += 2)
    {
        # Převést cílové kódy na řetězec.
        my $cil = $_[$i+1] eq "" ? "" : join("", map {chr($_)} split(/\+/, $_[$i+1]));
        $tabulka->{$_[$i]} = $cil;
    }
    return $tabulka;
}



#------------------------------------------------------------------------------
# Přepíše text na základě převodní tabulky řetězců (např. z jednoho písma do
# druhého).
#------------------------------------------------------------------------------
sub prepsat_tabulka
{
    my $tabulka = shift; # odkaz na hash
    my $zdroj = shift; # řetězec, který má být přepsán
    my $cil; # přepsaný řetězec
    # Zjistit délku nejdelšího zdrojového řetězce. Na každé pozici pak budeme testovat prefixy až do této délky!
    my $maxdelka = 0;
    foreach my $klic (keys(%{$tabulka}))
    {
        my $delka = length($klic);
        if($delka>$maxdelka)
        {
            $maxdelka = $delka;
        }
    }
    # Projít řetězec a přepsat ho.
    while($zdroj)
    {
        # Zkusit všechny délky prefixů, začít od nejdelšího.
        my $nalezeno = 0;
        for(my $i = $maxdelka; $i>0; $i--)
        {
            # Podívat se na prefix délky $i a zjistit, jestli ho známe jako zdrojový řetězec.
            if($zdroj =~ m/^(.{$i})/s && exists($tabulka->{$1}))
            {
                # Nalezený prefix odstranit ze zdroje.
                $zdroj =~ s/^(.{$i})//s;
                # Jeho přepis přidat do cíle.
                $cil .= $tabulka->{$1};
                # Ukončit hledání, resp. začít ho v dalším kole znova od začátku.
                $nalezeno = 1;
                last;
            }
        }
        # I pokud nebyl rozpoznán žádný prefix, musíme užrat alespoň první znak, abychom se hnuli z místa.
        unless($nalezeno)
        {
            $zdroj =~ s/^(.)//s;
            $cil .= $1;
        }
    }
    return $cil;
}



#------------------------------------------------------------------------------
# Vrátí převodní tabulku jako text HTML, který lze využít jako nápovědu.
#------------------------------------------------------------------------------
sub prevest_tabulku_do_html
{
    my $tabulka = shift;
    my $html;
    $html .= "<table border='1'>\n";
    $html .= "  <tr>\n";
    $html .= "    ";
    my @klice = sort(keys(%{$tabulka}));
    foreach my $klic (@klice)
    {
        # Tabulka může obsahovat nesamostatné znaky.
        # Mezerami kolem nich se snažíme zajistit, aby se nekombinovaly se značkami HTML.
        $html .= "<td> $klic </td>";
    }
    $html .= "\n";
    $html .= "  </tr>\n";
    $html .= "  <tr>\n";
    $html .= "    ";
    foreach my $klic (@klice)
    {
        # Tabulka může obsahovat nesamostatné znaky.
        # Mezerami kolem nich se snažíme zajistit, aby se nekombinovaly se značkami HTML.
        $html .= "<td> $tabulka->{$klic} </td>";
    }
    $html .= "\n";
    $html .= "  </tr>\n";
    $html .= "</table>\n";
    return $html;
}



#------------------------------------------------------------------------------
# Vygeneruje nápovědu v HTML.
#------------------------------------------------------------------------------
sub napoveda
{
    my $html;
    $html .= "<h1>Nápověda</h1>\n";
    $html .= "<h2>Latinka s&nbsp;diakritikou</h2>\n";
    $html .= prevest_tabulku_do_html($tabulka_latin);
    $html .= "<h2>Arabština</h2>\n";
    $html .= '<tt>-(ara?b?)1-(.*?)-(?:\1)0-</tt>'."\n";
    $html .= prevest_tabulku_do_html($tabulka_arab);
    $html .= "<h2>Urdština</h2>\n";
    $html .= '<tt>-(urd)1-(.*?)-(?:\1)0-</tt>'."\n";
    $html .= prevest_tabulku_do_html($tabulka_urdu);
    $html .= "<h2>Dévanágarí</h2>\n";
    $html .= '<tt>-(dev)1-(.*?)-(?:\1)0-</tt>'."\n";
    $html .= prevest_tabulku_do_html($tabulka_devanagari);
    $html .= "<h2>Gurmukhí</h2>\n";
    $html .= '<tt>-(gur)1-(.*?)-(?:\1)0-</tt>'."\n";
    $html .= prevest_tabulku_do_html($tabulka_gurmukhi);
    $html .= "<h2>Telugština</h2>\n";
    $html .= '<tt>-(tel)1-(.*?)-(?:\1)0-</tt>'."\n";
    $html .= prevest_tabulku_do_html($tabulka_telugu);
    return $html;
}



#------------------------------------------------------------------------------
# Inicializuje převodní tabulky. Odkazy na tabulky jsou globální proměnné.
# Nemůže se ve zdrojáku nacházet před funkcemi, které volá.
#------------------------------------------------------------------------------
BEGIN
{
    $tabulka_latin = sestavit_tabulku_latin();
    $tabulka_arab = sestavit_tabulku_arab();
    $tabulka_urdu = sestavit_tabulku_urdu();
    $tabulka_devanagari = sestavit_tabulku_devanagari();
    $tabulka_gurmukhi = sestavit_tabulku_gurmukhi();
    $tabulka_telugu = sestavit_tabulku_telugu();
}



1;
